#define SLOW_MPI 0                               /* old version, obsolete */
#define MAX_out 1000        /* maximum lines for debugging/verbose output */
/*  UseAlpha: maximum number parameter alpha used
 *        0 = no parameter alpha (default)
 *        1 = set parameter  alpha[0]    was alpha in old version
 *        2 = set parameters alpha[0..1] was alpha, beta in old version
 *        3 = set parameters alpha[0..2] was alpha,beta,gamma (old version)
 */
#define UseAlpha      0
/* start angle: */
//#define StartAlpha    0.0
//#define StartAlpha    (M_PI/6.0)   /* triangle AF in external field 1 */
#define StartAlpha    (M_PI/2.0)     /* triangle AF in external field 2 */
#define StepAlpha     (0.001*M_PI)         /* step of alpha */
#define StopAlpha     (StartAlpha+2*M_PI)  /* stop angle    */
/*  SearchAlphaMin:
 *        0 = no search, alpha fixed (default)
 *        1 = search alpha[0] (was alpha pre0807)
 *        2 = search alpha[0..1] (was beta  pre0807)
 *        3 = search alpha[0..2] (was gamma pre0807)
 */
#define SearchAlphaMin 0
/* alpha dependency of terms: */ 
//#define USR_FILE  "sample_bipartite_external_field.cpp"          
//#define USR_FILE  "sample_bipartite_external_field_j_j_dash.cpp" 
//#define USR_FILE  "sample_triangle_external_field_1.cpp"
#  define USR_FILE  "sample_triangle_external_field_2.cpp"
// set UseExpAlpha equal to 0 if you don't want to use user defined file in find expectation values.
#define UseExpAlpha 0
// Jun08 Joerg: sample*expectation*.cpp added to sample*.cpp files
// Jun08 Joerg: alpha, beta, gamma replaced by alpha[0..2]

/*

  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:      CCM (solve.C)
  
  Written by: Damian J. J. Farnell

  Date:      Wed Mar 15 15:20:28 GMT 2000

  Updated:   Joerg Schulenburg 2004-2008

  ===================================================

  ---------------------------------------------------

  PROGRAM NAME: SOLVE.C 

  PURPOSE: Routine that solves saved CCM
           equations using newton-raphson
	   or direct iteration. Sublattice
           magnetisation determined and 
	   ground-state energy (+ derivatives).

  ---------------------------------------------------

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// #include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "Constants.h"
#include "Externals.h"
#include "FunctionPrototypes.h"
#include "nr.h"
#include "nrutil.h"
#include "storage.h"
#include "f2c.h"
#include "eispackd.h"

#if (PARALLEL_FLAG>0)
#include <mpi.h>
#endif

// char gs_name[100]="";
char input_name[200];
// char configurations[200];
int mapping[100]={0}; 
 /* The variable called "mapping" maps the 81 terms in the excited-state 
    equations to specific terms in the ground-state equations. 
    This is so that the "term" variable in the definition of the 
    delta[bonds][term] is correct. Again, seems to work quite well! */

/* Parameter array of size Params.bonds d1*H1+d2*H2+... */
// double (*delta)[62+1]=NULL;
// double (*dincr)[62+1]=NULL; // increment to next delta (derivation)
double delta[5][62+1]; // depends from usr_angle*usr_parameter=f(bond,term63)
double dincr[5][62+1]; // increment to next delta (derivation)

/* -------------------------------------------------------------
 *  array of parameters depending on one variable parameter   
 *  sample: d1=cos(d), d2=sin(d) with H= d1*H1+d2*H2         
 */
static int PFLINE= __LINE__ + 1;   // this is only a localizator
// Joerg: depends from alpha only? can it depend from term (term13 or term63)?
// Joerg Jun08: remove term dependence
// double usr_parameter(int bonds, int term, double alpha)
double usr_parameter(int bonds, double alpha)
{
    /* add your model specifique parameter here! H=sum_i u(i,alpha)*H_i */
#if 1
  // square J1 J2 or xxzSquare
  //  tdelta * H1 + H2 + H3 + ...
  if (bonds==1) return alpha;
#elif 1
  // condition for triangular AF in external field: both model states
  if (bonds==1 || bonds==2) return alpha;
#else
  //  cos(alpha) * H1 + sin(alpha) * H2 + H3 + ...
  if (bonds==1) return cos(alpha*M_PI);
  if (bonds==2) return sin(alpha*M_PI); 
  if (bonds==3) return 1; 
#endif
  return 1.0;
}

#if UseAlpha
// include user defined function usr_angle(bond,term13,*alpha,numalpha)
// Jun08 Joerg: includes USR_EXP_FILE in USR_FILE now (more simple)
#include USR_FILE  
#endif

double findEnergy(double *x);
double findExcitations(double *x, double *bra_x, double *alpha);
int term_get3(int *descriptors, double *multiplier); // ToDo: remove
void sublattice_term(int length, int (*configuration)[5], int *subTerm);
double findStagMag(double *x, double *bra_x, double *subLatticeMag);
void findBraState(double *x, double *bra_x);
void findBraStateRHS(double *rhs, double *x);
void findBraEquations(double **m, double *x);
int directIteration(int maximumNumIterations, int numConfigs, 
		    double tolerance, int &iterations, double *oldX);
int directIterationBraState(int maximumNumIterations, int numConfigs, 
			    double tolerance, int &iterations, 
			    double *oldX, double *oldBraX);

void solve(void)
{
  int i, j, k;
  double *x, *bra_x;
  // char root_name[200], append[]=".dat";
  char output[200]; // , output2[200];
  //  char path[]="../Data/";
  //  char v_j2[]="_v_j2";
  //  int  **configuration, length;
  int iterations, iterations2;
  double energy=0;
  //  double energy2;
  //  double specHeatA, specHeatC;
  int status; // , stat;
  int sign = 1;

  double stagMag;
  /* we need it to find the inflection points of E(p) curves */
  double Energy1, Energy2;
  double dEnergy, ddEnergy;
  double stagMag1, stagMag2;
  double dStagMag, ddStagMag;
  double alpha[3]={0.0, 0.0, 0.0}; // alpha, beta, gamma
  double excitationEnergy=0.0;
  
  double start, stop, tdelta;
#if UseAlpha
  double d_alpha[3], old_Energy[3], minE=1e9, a_minE=0;
#endif

  double subLatticeMag[MAX_UNIT_CELL+3]={0.0};

  double *g_exp;
  g_exp=      dvector(0,Params.nExpect+5);

  // configuration=imatrix(0,MAX_LSUBM+1,0,5);

  time_t theTime, t0, t1;
  
  // now get the mappings right for the delta terms DJJF Jun08 Excitation
  //   map 1..81 to 1..62 [ to 1..13?]
  // see Externals.h map81[81+1] to 1..13 (remove mapping[81+1] to 1..62)
  
  for (i= 1;i<= 5;i++) mapping[i]= 2; // zz 1
  for (i= 6;i<= 6;i++) mapping[i]= 6; // z+ 2
  for (i= 7;i<=19;i++) mapping[i]= 8; // z- 3
  for (i=20;i<=20;i++) mapping[i]=14; // +z 4
  for (i=21;i<=33;i++) mapping[i]=16; // -z 5
  for (i=34;i<=36;i++) mapping[i]=22; // +- 6
  for (i=37;i<=39;i++) mapping[i]=24; // -+ 7 (no ++ 8)
  for (i=40;i<=73;i++) mapping[i]=27; // -- 9
  for (i=74;i<=74;i++) mapping[i]=54; // z  10
  for (i=75;i<=77;i++) mapping[i]=55; // -  11
  for (i=78;i<=81;i++) mapping[i]=58; // z^2 12 (no 13)
  for (i=1;i<=81;i++)
    if (map62[mapping[i]] != map81[i]) {
      printf("ERROR wrong mapping: map62[mapping[%d]]=%d != map81[%d]=%d\n",
              i, map62[mapping[i]], i, map81[i]);
    }
 
  //  solutionStrategy=1;
  //  momentum=0.7;         /* see Constants.h */

  if (PARALLEL_FLAG > 0 && solutionStrategy == 1){
    printf("Direct Iteration only for parallel processing...\n"
	   "Exiting!!\n\n");
    ERR(0);
  }

  time(&theTime);
  srand((int) theTime);

   strcpy(input_name,    Control.name); strcat(input_name,    ".terms");
   strcpy(output,        Control.name); strcat(output,        ".output");
// strcpy(configurations,Control.name); strcat(configurations,".configs");
// strcpy(gs_name,       Control.name); strcat(gs_name,       "_gs");

  if (myid==0 || myid==npe-1)
    printf(" solve: myid=%3d input=%s\n"
           "                output=%s\n",
           myid, input_name, output);

  // terms_fp=fopen(input_name,"r"); => opened in ccm.C now

  if (myid == 0) {
    fp2=fopen(output,"w");  
    if (!fp2) {
      printf("Cannot Open %s...\nExiting!!\n\n", output);
      ERR(1);
    }
  }

  // ground-state energy equation.

  // fp_gs=fopen(gs_name,"r");
  
  Energy2=1.0;    Energy1=2.0;   /* set initial values, no meaning */
  stagMag2=1.0;   stagMag1=2.0;

  x=      dvector(0,NumberConfigurations+5);
  bra_x=  dvector(0,NumberConfigurations+5);

  for(i=1;i<=NumberConfigurations;i++) { 
    x[i]=    0.000001; 
    bra_x[i]=0.000000001; 
  }
  // x[1]=0.1; // testwise, ToDo: failes, why?
  x[0]=    1.0; /* dummy index set to 1.0 for simplification, Joerg Sep05 */
  bra_x[0]=1.0; /* dummy index set to 1.0 for simplification, Joerg Sep05 */

  // x[2] = 0.064048;
  // x[3] = 0.000000;
  // x[4] = 0.000000;
  // x[5] = 0.064048;
  // x[6] = 0.064048;
 
  if (Control.output > -1 && myid == 0) {
    printf("============================================================\n\n");
    printf("SOLVE CCM EQUATIONS: START ALGORITHM.\n\n");
    printf("============================================================\n\n");
  }

  if (increment > 0.0) { start= minimum;  stop= maximum; }
    else               { start=-maximum;  stop=-minimum; }

   // Joerg Sep05: changing parameterfunction gives you more flexibility
  if (myid==0){ 
    printf("usr_parameter can be defined in "__FILE__" L%d\n",PFLINE);
    printf("solutionStrategy= %d (%s)\n",
            solutionStrategy,
          ((solutionStrategy==1)?"newton-raphson":"direct iteration"));
  }
  /* array of parameters depending on one variable Parameter
   *  sample: d1=cos(d), d2=sin(d) with H= d1*H1+d2*H2
   */ 
  // delta = malloc((Params.bonds+1)*sizeof(double)*63);
  // dincr = malloc((Params.bonds+1)*sizeof(double)*63);
  // if (!delta || !dincr) {
  //   printf("input.C: delta = malloc() failed...\nExit!\n");
  //   exit(1);
  // }
  
  sign = ((increment<0.0) ? -1 : +1);

#if UseAlpha
    alpha[0] =  StartAlpha;  // starting angle, 
    alpha[1] = -0.5*M_PI;
    alpha[2] =  0.5*M_PI;
    d_alpha[0] = StepAlpha;  // starting angle step
    d_alpha[1] = StepAlpha;  // starting angle step
    d_alpha[2] = StepAlpha;  // starting angle step
#endif

  // loop over different couplings J, go from known to unknown angle
  for(tdelta=start; tdelta<=stop; tdelta+=fabs(increment)) {

    // D1(Delta)*H1 + D2(Delta)*H2 + D3(Delta)*H3 + ... , Joerg Oct05
    if (myid==0)
    printf("\nModel Parameter Delta = %6.3f\n", sign*tdelta);

    for (int i=1; i<= Params.bonds; i++) {
      int ja=1; /* start range */ 
      for (j=1; j<=62; j++) {  // ToDo: use map62[1..62]
        delta[i][j] = usr_parameter(i,sign* tdelta );
        /* Joerg: dincr is only used for precondition (shift startvector) */
        dincr[i][j] = usr_parameter(i,sign*(tdelta+0.5*fabs(increment)))
                    - usr_parameter(i,sign*(tdelta-0.5*fabs(increment)));
        if (j>1 && delta[i][ja]!=delta[i][j] && myid==0) {
          printf(" Delta[bond=%d][term=%d...%d]= %6.3f\n",i,ja,j-1,delta[i][ja]);
          ja=j; /* next range */
        }
      }
      if (myid==0)
      printf(" Delta[bond=%d][term=%d...%d]= %6.3f\n",i,ja,j-1,delta[i][ja]);
    }

#if UseAlpha
    old_Energy[0] = 1.0e21; // biggest possible value
    // find the the minimum E(angle), we assume that exact one minimum exist
    // increase accuracy of angle from 1e-1 to 1e-9 (interval)
    // ToDo: that seems not a good algorithm, use Newton method at least
    // go down with const steps, if it goes up, return with smaller steps
    for (d_alpha[0]=StepAlpha;
      fabs(d_alpha[0])> 2.0e-9; alpha[0] += d_alpha[0]) {
	  old_Energy[1]=1.0e21;
       for (d_alpha[1]=fabs(d_alpha[0]); 
        fabs(d_alpha[1])> fabs(d_alpha[0])/4.0; alpha[1] += d_alpha[1]) {
	   old_Energy[2]=1.0e21;
       for (d_alpha[2]=fabs(d_alpha[0]);
        fabs(d_alpha[2])> fabs(d_alpha[0])/4.0; alpha[2] += d_alpha[2]) {
 
       // find the best angle (lowest energy) = angle +/- d_angle
       //
       // E                            5       6       7
       // step ------------------------1-------3---4-562------------- 
       // angle Pi/2                 +0.1   +0.125   +0.15

       for (int i=1; i<= Params.bonds; i++)
         for (j=1; j<=62; j++) {
           delta[i][j] = usr_angle(i,map62[j],alpha,UseAlpha )
                       * usr_parameter(i,sign* tdelta );
           dincr[i][j] = usr_parameter(i,sign*(tdelta+0.5*fabs(increment)))
                       - usr_parameter(i,sign*(tdelta-0.5*fabs(increment)));
           dincr[i][j]*=usr_angle(i,map62[j],alpha,UseAlpha );
           // dincr[i][j] = ??? 
         }
#endif
    iterations=0;
    iterations2=0;
    time(&t0);      // start time
    
    if (solutionStrategy==1) {
      
     // (a) newton-raphson ket state + LU decomp. bra state.

      status=mnewt((int) MAX_NUM_ITERATIONS, NumberConfigurations, 
		   (double) TOLERANCE, (double) TOLERANCE, iterations, x);
      
    } else {
      
      // (b) direct iterations for both ket and bra states.

      status=directIteration((int) MAX_NUM_ITERATIONS,
             NumberConfigurations, (double) TOLERANCE, 
	     iterations, x);

    }
    time(&t1);

    if (iterations >= MAX_NUM_ITERATIONS) {
       printf("WARNING: iterations=%d >= MAX_NUM_ITERATIONS\n", iterations);
       continue; // Mar08
       // break;
    }
    if (status<0 && !myid) {
       printf("WARNING: no solution found, status=%d, bad momentum?\n", status);
       ERR(1);
    }
    
    energy=findEnergy(x);
    if (Control.output>0 && myid==0) // print out later?
        printf("Eg= %18.14f it= %4d  Eg.i100t= %6.2fm\n",
           energy, iterations, (t1-t0)/.6/iterations);
    
#if UseAlpha
      if (Control.output>0 && myid==0)
        printf("Eg= %18.14f alpha/Pi= % 10.8f d_alpha/M_PI=% 9.1e\n",
               energy, alpha[0]/M_PI, d_alpha[0]/M_PI);
#if SearchAlphaMin >= 3
      if (Control.output>0 && myid==0)
         printf(" : gamma/Pi = % 10.8f : d_beta=% 10.2e\n",
	           alpha[2]/M_PI, d_alpha[2]);
      if (energy > old_Energy[2]) d_alpha[2] = -0.5 * d_alpha[2]; // go back
#else
      d_alpha[2] = 0;                            /* stop condition */
#endif
		 
      old_Energy[2] = energy; // last value
     } /* d_alpha[2] == d_gamma */

#if SearchAlphaMin >= 2
      if (Control.output>0 && myid==0)
         printf(" : beta/Pi = % 10.8f : d_beta=% 10.2e\n",
	           alpha[1]/M_PI, d_alpha[1]);
      if (energy > old_Energy[1]) d_alpha[1] = -0.5 * d_alpha[1]; // go back
#else
      d_alpha[1] = 0;                            /* stop condition */
#endif
		 
      old_Energy[1] = energy; // last value
     } /* d_alpha[1] == d_beta */

#if SearchAlphaMin
      if (energy > old_Energy[0]) d_alpha[0] = -0.5 * d_alpha[0]; // go back
#else
      d_alpha[0] = 0;                            /* stop condition */
#endif
      old_Energy[0] = energy; // last value
      if (energy<minE && iterations<MAX_NUM_ITERATIONS) { // Mar08
        minE=energy; a_minE=alpha[0];
      } 
    } /* d_alpha[0] */
    
    // alpha=a_minE; energy=minE; // ???
    if (Control.output>-1 && myid==0)
        printf("E(alpha)_lowest: Eg = %12.8f : alpha/Pi = % 10.8f\n",
               minE, a_minE/M_PI); // Mar08
    if (Control.output>-1 && myid==0)
        printf("E(alpha)_last:   Eg = %12.8f : alpha/Pi = % 10.8f\n",
               energy, alpha[0]/M_PI); // Mar08
#if SearchAlphaMin >= 3
    if (Control.output>-1 && myid==0)
        printf("E(gamma)_lowest: Eg = %12.8f : gamma/Pi = % 12.8f\n",
               energy, alpha[2]/M_PI);
#endif
#if SearchAlphaMin >= 2
    if (Control.output>-1 && myid==0)
        printf("E(beta)_lowest: Eg = %12.8f : beta/Pi = % 12.8f\n",
               energy, alpha[1]/M_PI);
#endif

#endif

    if (iterations >= MAX_NUM_ITERATIONS) continue;
      // break;

    time(&t0);

    if (solutionStrategy==1) {
      
     // (a) LU decomp. bra state.

      findBraState(x, bra_x);
      iterations2=100; // ???

    } else {
      
      // (b) direct iterations for bra states.

      status=directIterationBraState((int) MAX_NUM_ITERATIONS, 
	     NumberConfigurations, (double) TOLERANCE, 
	     iterations2, x, bra_x); 
    }

    time(&t1);

    // (c) the (sub)lattice magnetisation.
      
    stagMag=findStagMag(x,bra_x,subLatticeMag);
	
    if (Control.output>0 && myid==0) // print out later?
        printf(" M= %18.14f it= %4d  Mg.i100t= %6.2fm\n",
           stagMag, iterations2, (t1-t0)/.6/iterations2);
    
	
    // now find the generalised expectations.
	
    if (Control.greatExpectations) findGenExp(x,bra_x,g_exp,alpha);

    excitationEnergy=0.0;
    if (Control.excitations)
      excitationEnergy=findExcitations(x,bra_x,alpha);

    // PRINT OUT INFORMATION.
    
    // we need at least two predecessing steps
    if ( fabs(tdelta-start) <= fabs(1.1*increment) ) {
      dEnergy  = ddEnergy  = 0;
      dStagMag = ddStagMag = 0;
    } else {
      dEnergy  = (energy-Energy1)/(increment);
      ddEnergy = (energy-2*Energy1+Energy2)/(increment*increment);
      dStagMag = (stagMag-stagMag1)/(increment);
      ddStagMag= (stagMag-2*stagMag1+stagMag2)/(increment*increment);
      if (Control.output > 0 && myid == 0) {
        printf("Computational 1st Derivative Eg wrt delta = % f\n",dEnergy);
        printf("Computational 2nd Derivative Eg wrt delta = % f\n",ddEnergy);
        printf("Computational 1st Derivative  M wrt delta = % f\n",dStagMag);
        printf("Computational 2nd Derivative  M wrt delta = % f\n",ddStagMag);
      }
    }    
    // = -d^2 E_g/ d Delta^2 via computational evaluation only.
    
    Energy2=Energy1;      Energy1=energy;   // save older values
    stagMag2=stagMag1;    stagMag1=stagMag;

    if (Control.output > -1 && myid == 0) {
      printf("Delta = %6.4f : Eg = %12.8f : M = %12.8f : -dEg = %12.8f : Heat = %12.8f\n"
             "  x[1]=%12.9f : iterations = %4d\n"
             "bra[1]=%12.9f : iterations = %4d\n",
	     sign*tdelta, energy, stagMag, -dEnergy, ddEnergy,
             x[1], iterations, bra_x[1], iterations2);
      if(UnitCell.number>1) {
	for(k=1;k<=UnitCell.number;k++)
	  printf("Mag.[%d] = %12.8f\n", k, subLatticeMag[k]);
    }
#if UseAlpha
  printf("Alpha=%16.12f\n",alpha[0]/M_PI);
#endif
#if SearchAlphaMin >= 2
  printf(" Beta=%16.12f\n",alpha[1]/M_PI);
#endif
#if SearchAlphaMin >= 3
  printf("Gamma=%16.12f\n",alpha[2]/M_PI);
#endif
	// printf("\n");
      
  if (Control.excitations && myid==0)  // DJJF Jun08
    printf("Excited-State Energy =%16.10f\n", excitationEnergy);

  if (Control.greatExpectations>0) {
    double sum_g_exp=0;
    for (k=1;k<=Params.nExpect;k++) {
      printf("Generalised Expectation Value [%d] = %12.8f\n", k, g_exp[k]);
      sum_g_exp += g_exp[k];
    }
    // Jun08 Joerg: old was g_exp[1]+...+g_exp[3] which can be uninitialized
    printf("Lattice Magnetisation Triangular AF=%12.8f\n",sum_g_exp);
  }
  if (Control.output > 0) {
      int i, j0, j1;
      for(j0=j1=0,i=1;i<=NumberConfigurations;i++) {
        if (x[i]<0) j0++; if(bra_x[i]<0) j1++;
      }
      printf("\nnumber negativ coefficients (x,bra_x) = %d %d"
               "  NumConf= %d\n", j0, j1, NumberConfigurations);
      if (j0<100 && Control.output>1) { // lot of output
          for(j0=0,i=1;i<=NumberConfigurations;i++) {
            if (x[i]<0) {
              if(j0<1000) printf(" %d",i);
              if(j0==1000) printf("\n");
              j0++;
            }
          }
          printf("\n");
          for(j0=0,i=1;i<=NumberConfigurations;i++) {
            if (bra_x[i]<0) {
              if(j0<1000) printf(" %d",i);
              if(j0==1000) printf("\n");
              j0++;
            }
          }
          printf("\n");
        }
      }
      if (Control.output > 1) {
        int i; double sum=0,sumb=0,prod=0;
        for(i=1;i<=NumberConfigurations;i++) {
          if (i<Control.output*20)
            printf(" i x[i] bra[i] %4d % .9f % .9f\n",
                     i,x[i],bra_x[i]);
          if (i==Control.output*20) printf("...\n");
          sum+=x[i]*x[i]; sumb+=bra_x[i]*bra_x[i]; prod+=x[i]*bra_x[i];
        }
        printf(" x.x bra.bra x.bra  % .9f % .9f % .9f\n",sum,sumb,prod);
      }
    }

    if (myid == 0) {
      fprintf(fp2,"%16.12f %16.12f %16.12f %16.12f %16.12f %16.12f %d %d",
	sign*tdelta,energy,stagMag,-dEnergy,ddEnergy,x[1],iterations,iterations2);
#if UseAlpha
  fprintf(fp2,"%16.12f ",alpha[0]/M_PI);
#endif
#if SearchAlphaMin >= 2
  fprintf(fp2,"%16.12f ",alpha[1]/M_PI);
#endif
#if SearchAlphaMin >= 3
  fprintf(fp2,"%16.12f ",alpha[2]/M_PI);
#endif
      if(UnitCell.number>1) {
	for(k=1;k<=UnitCell.number;k++)
	  fprintf(fp2,"%16.12f ",subLatticeMag[k]);
      }
      double sum1=0.0;
      if (Control.greatExpectations) {
        for (k=1;k<=Params.nExpect;k++) sum1+=g_exp[k];
        for (k=1;k<=Params.nExpect;k++) fprintf(fp2,"%16.12f ", g_exp[k]);
      }
      fprintf(fp2,"%16.12f ",sum1); // for triangular AD lattice mag only
      if (Control.excitations) { 
        fprintf(fp2,"excErg= %16.12f ", excitationEnergy);
      }
      fprintf(fp2,"\n");
      fflush(fp2);
    }
  }
  if (Control.output > -1 && myid == 0) {
    printf("============================================================\n\n");
    printf("SOLVE CCM EQUATIONS: SUCCESSFULLY COMPLETED!\n\n");
    printf("============================================================\n\n");
  }

  // fclose(terms_fp);    terms_fp=NULL;
  if (myid==0) fclose(fp2);

 // ferase(fp_exp);

  // stat=remove(gs_name);
  // stat=remove(input_name);  
  // stat=remove(configurations);

}

/* called by solve() and directIteration(only_printout)
 * new: x[0]=1 now (Sep05)
 * 
 */
double findEnergy(double *x) 
{

  int i, bonds;
  int indices[7]={0};
  double dummy;
  double sum;

  sum=0.0; // nio=7;
  
  term_rewind(0);

  for(bonds=1;bonds<=Params.bonds;bonds++) {

    for(i=1;;i++) {
      
      term_get(0,indices,&dummy);  // get gse_terms
      
      if (indices[0]<0 || indices[0] > NumberConfigurations) break;
      
      // Joerg Sep05:
      //   according to gse_terms only indices[2..3] can be nonzero
      // gse_term{3}      : (i2,i3) = two single-site configs
      // gse_term{2,4,5,7}: (i2, 0) = two-site neighbour config
      // gse_term{1,6,8}  : ( 0, 0)
      // squareLSUB10 -> B1: (0,1,0,0,0,0,-0.5), B2: (0,27,1,0,0,0,-1.0)
      sum += dummy * x[indices[2]]
                   * x[indices[3]] * delta[bonds][indices[1]];
    }

  }
  return sum;
}
 
// in/out: oldX - startvector + iterationvector
// in(out: dx   - difference oldX(out) - oldX(in)  (only if precondition)
//                used to modify start vector (speed up?) -> removed
int directIteration(int maximumNumIterations, int numConfigs, 
		    double tolerance, int &iterations, double *oldX)
{  
  int i, j, bonds, max_err_j[3];
  int indices[7]={0};
  double dummy;
  double xi2, xi3, xi4, xi5;
  double *lhs, *rhs,    /* Joerg: left-hand side, right hand side ? */ 
         min_error, error, mom, err;
  double max_err[3]; // store 3 biggest changes
//  double tempSum1, sum1, tempSum2, sum2;
//  double *tempLHS,  *tempRHS;
//  double *tempLHS2, *tempRHS2;

  error = 0.0; // JS Nov05
  min_error = 10.0; // JS Apr06, to stop divergence problems LSUB14 chain
  mom=momentum;     // default is 0.7 set in Constants.h 

  if (myid==0)
    printf("Direct iteration ket state (tolerance=%g, momentum=%g):\n",
            tolerance, momentum);

  lhs= dvector(0,numConfigs+5);
  rhs= dvector(0,numConfigs+5);

  for(i=1;i<=numConfigs;i++) {
    lhs[i]=0.0;
    rhs[i]=0.0;
  }
  
  /* Starting values */
  if (Control.output > 0 && myid == 0) 
      printf(" iteration= %4d  energy= %16.10f  oldX[1]= %16.10f\n",
	     0, findEnergy(oldX), oldX[1]);

  for(iterations=1; iterations<=maximumNumIterations; iterations++) {
    
    term_rewind(1);

    /* calculate rhs and lhs from input-file */
    for(bonds=1; bonds<=Params.bonds; bonds++) {
      
      /* why loop over bonds? inner loop does not depend on bonds? */
      for(i=1;;i++) {
	
        term_get(1,indices,&dummy);
	
	if (indices[0] < 0 || indices[0] > NumberConfigurations)
	  break;
	
        /* now: oldX[0]=1.0, Joerg Sep05 */
        xi2=oldX[indices[2]]; /* pointing to a configuration index or 0 */
        xi3=oldX[indices[3]];
        xi4=oldX[indices[4]];
        xi5=oldX[indices[5]]; /* 4 indices only for term34 */
    
        /* ToDo: indices are sorted, so that zeros are at end (indices[5]) */
        /* Joerg: we look for terms where indices[2]==indices[0] and rest 0 */
        if (indices[3]==0 && indices[2]==indices[0])
          /* Joerg Nov05: diagonal term? called rare */
          lhs[indices[0]]-=dummy*delta[bonds][indices[1]];
        else { /* called often */
          /* Joerg Nov05: offdiagonal terms? */
          rhs[indices[0]]+=dummy*xi2*xi3*xi4*xi5
                                *delta[bonds][indices[1]];
        }
      }
    }

    error=0.0;

    // only DEBUG (myid1=1 myid2=2 ... myid0=npe-1)
    for(j=((myid==0)?npe:myid);j<=numConfigs;j+=npe) {  // parallel 1...
      /* old: using tolerance=1e-12 here is not a good idea!? */
      if (fabs(lhs[j]) < 1.0e-9) {  // usually -0.5,-1.0,...
	printf("Error in direct iterations: no left-hand side...\n"
          "it= %4d  cfg= %4d  lhs= % .4g\n"
	  "Exiting!!\n\n", iterations, j, lhs[j]);
        ERR(1);
      }
    }
#if (PARALLEL_FLAG> 0) 
    /* MPI_Barrier(MPI_COMM_WORLD); dont needed!? */
    /* new faster code by JS */
    /* only every npe-th element is nonzero, summarize */
    // we use lhs temporary to sum over threads
    for(j=1;j<=numConfigs;j++) {   /* [0] is not used */
      if ((j%npe) == myid) lhs[j]=rhs[j]/lhs[j]; else lhs[j]=0.0;
    }
    // store the sum of lhs to rhs
    MPI_Allreduce(&lhs[1], &rhs[1], numConfigs,
                  MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    for(j=1;j<=numConfigs;j++) lhs[j]=1.0; // parallel-to-single adaption
    // now we can use single-task algorithm to go further
#endif

    max_err[0]=max_err[1]=max_err[2]=0; 
    max_err_j[0]=max_err_j[1]=max_err_j[2]=0;
    for(j=1;j<=numConfigs;j++) {
      error+=fabs(err=oldX[j]-rhs[j]/lhs[j]);
      // store 3 biggest changes
      if (fabs(err)>fabs(max_err[0])) {
         max_err[2]=max_err[1]; max_err_j[2]=max_err_j[1];
         max_err[1]=max_err[0]; max_err_j[1]=max_err_j[0];
         max_err[0]=err; max_err_j[0]=j; 
      } else
      if (fabs(err)>fabs(max_err[1])) {
         max_err[2]=max_err[1]; max_err_j[2]=max_err_j[1];
         max_err[1]=err; max_err_j[1]=j; 
      } else
      if (fabs(err)>fabs(max_err[2])) {
         max_err[2]=err; max_err_j[2]=j; 
      }
      // ToDo: get j of max(err)
    }
    // error = fabs(max_err[0]); // other |X|=distance definition
#if 0  // can generate lot of iterations
    if (error > 2*min_error && iterations>10) {   // should go down allways
      mom = 1.0 - 0.5*(1.0-mom);
      min_error = error;
      if (myid==0)
        printf(" It=%d: reduce 1-momentum to %g to ensure convergence\n",
             iterations, 1-mom);
    }
#endif
    // max_X_j=0; max_X=0.0; // only DEBUG
    for(j=1;j<=numConfigs;j++) {
      oldX[j]=mom*oldX[j]+(1.0-mom)*rhs[j]/lhs[j]; // some parallel overhead
      lhs[j]=0.0;
      rhs[j]=0.0;
      // ToDo: outside It.: if (fabs(oldX[j])>fabs(max_X)) {max_X=oldX[j]; max_X_j=j;}
    }
    
    /* calling findEnergy() here is a vast of time ??? (Joerg) */
    if (Control.output > 0 && myid == 0 && ((iterations % 10)==0 
      || Control.output> 2 || error > 0.0625 || min_error == error)) {
      // err= sum(fabs(dX[i]),i) (i,dX[i])_for_max(fabs(dX[i])
      printf(" %4d energy= %16.10f err= %9.2e %4d %9.2e %4d %9.2e\n",
	     iterations, findEnergy(oldX), error,
             max_err_j[0], max_err[0], max_err_j[1], max_err[1]);
    }

    if (isnan(error)) {   /* break if +-Inf or NaN */
       printf(" iteration = %4d : energy = %16.10f : "
              "error = %14.12f\nError directIteration: NaN-BREAK\n",
              iterations, findEnergy(oldX), error);
       iterations=maximumNumIterations; break;
    }

    if (min_error > error) min_error = error;
    // min_error=fabs(max_err[0]); mom=momentum;
    if (error < tolerance) break; // convergence
    if (error > 10*min_error && iterations>99) { iterations=-1; break; } // divergence

  }

  if (iterations>=maximumNumIterations) {
    // if (!myid)
    ERR("directIteration Error: iterations>=maximumNumIterations"); // exit
  }

  free_dvector(lhs,0,numConfigs+5);
  free_dvector(rhs,0,numConfigs+5);

  return iterations;
}

// in: x[NumConf]
// out: fvec[NumConf], fjac[NumConf][NumConf]
// called by mnewt()=NewtonRawson in num_methods.C
void usrfun(int n, double *fvec, double **fjac, double *x, int iterations)
{  
  int i, k, bonds;
  int indices[7]={0};
  double dummy;
  double xi2, xi3, xi4, xi5, multiplier;
  static int warn1=0;
  
  term_rewind(1);
  
  if (sizeof(int *)<=4 && NumberConfigurations>=16384 && myid==0 && warn1==0){
    /* may be the 2D-array is to big for a 32bit-system (nC>=(2GB/8)^.5) */
    printf("Warning usrfun: array may be to big for a 32bit-system\n");
  }
  for(i=1;i<=NumberConfigurations;i++) {    
    fvec[i]=0.0;
    for(k=1;k<=NumberConfigurations;k++) 
      fjac[i][k]=0.0;  /* may fail if nC*nC*double>2GB */
  }  
 
  i=0;
  for(bonds=1;bonds<=Params.bonds;bonds++) {

    for(;;i++) {
      
      term_get(1,indices,&dummy);
      
      if (indices[0] < 0 || indices[0] > NumberConfigurations)
	break;
      
      x[0]=1.0; /* obsolete??? Joerg Sep05 */
      xi2=x[indices[2]];
      xi3=x[indices[3]];
      xi4=x[indices[4]];
      xi5=x[indices[5]]; 

      multiplier = delta[bonds][indices[1]]*dummy;
      
      fvec[indices[0]] += xi2*xi3*xi4*xi5*multiplier;
      
      if (indices[2] > 0) fjac[indices[0]][indices[2]] += xi3*xi4*xi5*multiplier;
      if (indices[3] > 0) fjac[indices[0]][indices[3]] += xi2*xi4*xi5*multiplier;
      if (indices[4] > 0) fjac[indices[0]][indices[4]] += xi2*xi3*xi5*multiplier;
      if (indices[5] > 0) fjac[indices[0]][indices[5]] += xi2*xi3*xi4*multiplier;
       
    }
  }

  if (Control.output > 10 && myid == 0) {
    printf("usrfun (read %d):\n",i);
    for (i=1; i<=NumberConfigurations; i++) {
      if (i==6 && NumberConfigurations>10) printf (" ...\n");
      if (i>5 && i<NumberConfigurations-4) continue;
      for (k=1; k<=NumberConfigurations; k++) {
        if (k<=5 || k>NumberConfigurations-5) printf(" % 10.8f",fjac[i][k]);
        if (k==6 && NumberConfigurations >10) printf (" ...");
      }
      printf(" : % 10.8f \n",fvec[i]);
    }
  }
  
}

/*****************************************************
 *                                                   *
 * Obtain bra state (LU decomp.)                     *
 *                                                   *
 *****************************************************
 *                                                   *
 * x_I        -- ket-state correlatin coefficient.   *
 * \tilde x_I -- bra-state correlatin coefficient.   *
 *                                                   *
 * H = e_0 + sum_I \tilde x_I F_I                    *
 *                                                   *
 *****************************************************
 *                                                   *
 * d H / d (\tilde x_I) =>                           *
 *                                                   *
 * F_I = 0    <= ket-state equations                 *
 *                                                   *
 *****************************************************
 *                                                   *
 * d H / d x_I =>                                    *
 *                                                   *
 * d(e_o)/d(x_I)+sum_J \tilde x_J (d(F_J)/d(x_I)) =0 * 
 *            <= bra-state equations                 *
 *                                                   *
 * use LU decomposition to solve the equations:      *
 *                                                   *
 * F(I,J) (\tilde X(J)) = - e_o(I)                   *
 *                                                   *
 *****************************************************/

void findBraState(double *x, double *bra_x)
{

  /* Determines the bra-state correlation coefficients
     using LU decomposition.
     */

  int      i;
//  int      status;
  void     lubksb(double **a, int n, int *indx, double b[]);
  void     ludcmp(double **a, int n, int *indx, double *d);
  int      *ind;
  double   d;
  double   **bra_eqn;
  // int      length;
  // int      iterations;

  for(i=1;i<=NumberConfigurations;i++) 
    bra_x[i]=0.0; 

  ind=ivector(0,NumberConfigurations+1);
  bra_eqn=dmatrix(0,NumberConfigurations+1,0,NumberConfigurations+1); 

  findBraStateRHS(bra_x,x); // bra_x (-=) terms(x)
  
  // find sum_J \tilde x_J (d(F_J)/d(x_I)) == F(I,J)
  
  findBraEquations(bra_eqn, x);
  
  // LU decomp Numerical Recipes to solve: F(I,J) (\tilde X(J)) = - e_o(I)
  
  ludcmp(bra_eqn,NumberConfigurations,ind,&d);
  lubksb(bra_eqn,NumberConfigurations,ind,bra_x);
  
  free_ivector(ind,0,NumberConfigurations+1);
  free_dmatrix(bra_eqn,0,NumberConfigurations+1,0,NumberConfigurations+1);

}
  

/*****************************************************
 *                                                   *
 * Obtain bra state (direct iteration)               *
 *                                                   *
 *****************************************************
 *                                                   *
 * x_I        -- ket-state correlatin coefficient.   *
 * \tilde x_I -- bra-state correlatin coefficient.   *
 *                                                   *
 * H = e_0 + sum_I \tilde x_I F_I                    *
 *                                                   *
 *****************************************************
 *                                                   *
 * d H / d (\tilde x_I) =>                           *
 *                                                   *
 * F_I = 0    <= ket-state equations                 *
 *                                                   *
 *****************************************************
 *                                                   *
 * d H / d x_I =>                                    *
 *                                                   *
 * d(e_o)/d(x_I)+sum_J \tilde x_J (d(F_J)/d(x_I)) =0 * 
 *            <= bra-state equations                 *
 *                                                   *
 * Use direct iteration to solve:                    *
 *                                                   *
 *    (\tilde X(I)) = { - e_o(I) - \sum_{J != I}     *
 *                F(I,J) (\tilde X(J))}/F(I,I)       *
 *                                                   *
 * for the \tilde X(I)'s.                            *
 *                                                   *
 *****************************************************/

int directIterationBraState(int maximumNumIterations, int numConfigs, 
			    double tolerance, int &iterations, 
			    double *oldX, double *oldBraX)
{

  double error, min_error, mom;
  
  int i, j, bonds;
  int indices[7]={0};
  double dummy[1];
  double xi2, xi3, xi4, xi5,multiplier;
//  double tempSum1, sum1,tempSum2, sum2;
  double subLatticeMag[MAX_UNIT_CELL+3];

  double *lhs, *rhs;
#if (PARALLEL_FLAG> 0)
  double *tempLHS, *tempRHS;
  tempLHS=dvector(0,numConfigs-1);
  tempRHS=dvector(0,numConfigs-1);
#endif
  
  error = 0.0;
  min_error= 10.0;
  mom = momentum;
  // mom = 0.7;  // 0.25
  if (myid == 0) printf("Direct iteration bra state (mom=%g):\n",mom);

  lhs=dvector(0,numConfigs+5);
  rhs=dvector(0,numConfigs+5);

  for(i=1;i<=numConfigs;i++) {
    lhs[i]=0.0;
    rhs[i]=0.0;
  }

  if (Control.output > 0 && myid == 0) 
    printf(" %4d  mag= %12.10f  braX[1]= %12.10f\n",
      0,findStagMag(oldX,oldBraX,subLatticeMag),oldBraX[1]);

  for(iterations=1;iterations<=maximumNumIterations;iterations++) {
        
#if (PARALLEL_FLAG == 0) 
    findBraStateRHS(rhs,oldX); // rhs (-=) terms(oldX)
#endif

    term_rewind(1);

    for(bonds=1;bonds<=Params.bonds;bonds++) {
      
      for(i=1;;i++) {
	
        term_get(1,indices,dummy);
	
	if (indices[0] < 0 || indices[0] > NumberConfigurations)
	  break;

        multiplier=delta[bonds][indices[1]];

	
	if (indices[2]==0) { // all indices are equal to 0
	  
	  // rhs[indices[0]]+=dummy[0]*multiplier;
	  
	} else {
	  
          oldX[0]=1.0;
	  xi2=oldX[indices[2]];
	  xi3=oldX[indices[3]];
	  xi4=oldX[indices[4]];
	  xi5=oldX[indices[5]];  /* only for term34 */

          if (indices[5]>0) {
            rhs[indices[5]]-=dummy[0]*xi2*xi3*xi4*multiplier*
                oldBraX[indices[0]];
          }

          if (indices[4]>0) {
              rhs[indices[4]]-=dummy[0]*xi2*xi3*xi5*multiplier*
                oldBraX[indices[0]];
          }

          if (indices[3]>0) {
              rhs[indices[3]]-=dummy[0]*xi2*xi4*xi5*multiplier*
                oldBraX[indices[0]];
          }

	  if (indices[2]>0) {
	    if (indices[2]==indices[0]  && indices[3] ==0) // [4..5] == 0 too
	      lhs[indices[0]]+=dummy[0]*xi3*xi4*xi5*multiplier;
	    else 
	      rhs[indices[2]]-=dummy[0]*xi3*xi4*xi5*multiplier*
		oldBraX[indices[0]];
	  }
	}
      }
    }
      
#if (PARALLEL_FLAG> 0)
    /* ToDo JS06: speedup if rhs/lhs or only error[npe] summed,
     *    changing oldX can be distributed over threads */
    MPI_Allreduce(lhs+1, tempLHS, numConfigs, 
		  MPI_DOUBLE, MPI_SUM,MPI_COMM_WORLD);
    MPI_Allreduce(rhs+1, tempRHS, numConfigs, 
		  MPI_DOUBLE, MPI_SUM,MPI_COMM_WORLD);
    for(i=1;i<=numConfigs;i++) {
      lhs[i]=tempLHS[i-1];
      rhs[i]=tempRHS[i-1];
    }
    findBraStateRHS(rhs,oldX);  // rhs (-=) terms(oldX)
#endif

    error=0.0;
    for(j=1;j<=numConfigs;j++) {
      if (fabs(lhs[j]) < 1e-9) {
	printf("Error in direct iterations bra state: no left-hand side...\n"
	       "Exiting!!\n\n");
	break;
      }
      error+=fabs(oldBraX[j]-rhs[j]/lhs[j]);
    }
    if (j<=numConfigs) break;
#if 0  // can generate lot of iterations
    if (error>8*min_error) { // New JS Apr06, 2* => 8* Mar08, ToDo: make bigger?
       mom = 1.0 - 0.5*(1.0 - mom);  /* 1-0.3 -> 1-0.15 -> 1-0.075 -> ... */
       min_error = error;
       if (!myid)
         printf(" It=%d: reduce 1-momentum to %g to ensure convergence\n",
                iterations, 1.0-mom);
       if (1.0-mom<1e-6) break;
    }
#endif
    for(j=1;j<=numConfigs;j++) {
      oldBraX[j]=mom*oldBraX[j]+(1.0-mom)*rhs[j]/lhs[j];
      lhs[j]=0.0;
      rhs[j]=0.0;
    }

    if (Control.output > 0 && myid == 0
       && ((iterations % 20)==0 || Control.output > 2)) 
      printf(" %4d  mag= %12.10f err= %9.2e %9.2e\n",
	iterations, findStagMag(oldX,oldBraX,subLatticeMag),
        error, error-min_error);

    if (error < min_error) min_error = error;
    if (error < tolerance) break;
    // NaN happens for j1-j2-neel j2>0.55
    if ((isnan(subLatticeMag[1]) || isnan(error))) ERR("NaN");
    if (error > 1e6) { iterations=-1; break; }       // divergence

  }

  if (iterations>=maximumNumIterations) {
    printf("directIterationBraState Error: i>=maximumNumIterations\n");
    ERR(0);
  }

  free_dvector(lhs,0,numConfigs+5);
  free_dvector(rhs,0,numConfigs+5);
#if (PARALLEL_FLAG> 0)
  free_dvector(tempLHS,0,numConfigs-1);
  free_dvector(tempRHS,0,numConfigs-1);    
#endif

  return iterations;
}

// input:  rhs, x (global: terms, Params.bonds, NumberConfigurations)
// output: rhs
// --- called from each thread ---
void findBraStateRHS(double *rhs, double *x) 
{
  int i, bonds;
  int indices[7]={0};
  double dummy[1];
  double xi2, xi3, xi4, xi5, multiplier;

  term_rewind(0);

  for(bonds=1;bonds<=Params.bonds;bonds++) {

    for(i=1;;i++) {
      
      term_get(0,indices,dummy);
      
      if (indices[0] < 0 || indices[0] > NumberConfigurations) break;
      
      multiplier=delta[bonds][indices[1]];

      // ToDo: how often? can be removed?
      if ((indices[2]==0 && indices[3]==0
        && indices[4]==0 && indices[5]==0)) continue;
	
        x[0]=1;
        xi2=x[indices[2]];
	xi3=x[indices[3]];
	xi4=x[indices[4]];
	xi5=x[indices[5]];
	
	if (indices[2] > 0) // ToDo: faster if using dummy rhs[0] ??? 
	  rhs[indices[2]]-= dummy[0]*xi3*xi4*xi5*multiplier;
	
	if (indices[3] > 0) 
          rhs[indices[3]]-= dummy[0]*xi2*xi4*xi5*multiplier;
	
	if (indices[4] > 0)
	  rhs[indices[4]]-= dummy[0]*xi2*xi3*xi5*multiplier;
	
	if (indices[5] > 0)
	  rhs[indices[5]]-= dummy[0]*xi2*xi3*xi4*multiplier;
    }
  }
}

void findBraEquations(double **m, double *x) 
{
  /* FIND BRA EQUATIONS: finds the Jacobian in order to determine 
   the bra state equations.
   */

  // find: sum_J \tilde x_J (d(F_J)/d(x_I))
  
  int i,k,bonds;
  int indices[7]={0};
  double dummy[1];
  double xi2, xi3, xi4, xi5, multiplier;
  
  term_rewind(1);
  
  for(i=1;i<=NumberConfigurations;i++)    
    for(k=1;k<=NumberConfigurations;k++) 
      m[i][k]=0.0;

  for(bonds=1;bonds<=Params.bonds;bonds++) {
    
    for(i=1;;i++) {
      
      term_get(1,indices,dummy);
      
      if (indices[0] < 0 || indices[0] > NumberConfigurations)
	break;
      
      multiplier=delta[bonds][indices[1]];

      if (!(indices[2]==0 && indices[3]==0 && indices[4]==0 && 
	    indices[5]==0)) {
	
        x[0]=1.0;
	xi2=x[indices[2]];
	xi3=x[indices[3]];
	xi4=x[indices[4]];
	xi5=x[indices[5]];
	
	if (indices[2] > 0)
	  m[indices[2]][indices[0]]+= dummy[0]*xi3*xi4*xi5*multiplier;
	
	if (indices[3] > 0)
	  m[indices[3]][indices[0]]+= dummy[0]*xi2*xi4*xi5*multiplier;
	
	if (indices[4] > 0)
	  m[indices[4]][indices[0]]+= dummy[0]*xi2*xi3*xi5*multiplier;
	
	if (indices[5] > 0)
	  m[indices[5]][indices[0]]+= dummy[0]*xi2*xi3*xi4*multiplier;
      }
    }
  }
}

/*****************************************************
 *                                                   *
 * Name: Staggered Magnetisation Routine.            *
 *                                                   *
 * Determines the lattice magnetisation once bra-    *
 * and ket-state equations have been solved.         *
 *****************************************************/

double findStagMag(double *x, double *bra_x, double *subLatticeMag)
{

  /* FIND STAG MAG: obtains the staggered magnetisation.
     */

  int      i, j;
  double   stagMag;
  int      length;
  int      (*configuration)[5];
  int      subTerm[MAX_UNIT_CELL+3];

  configuration = (int (*)[5]) malloc( (MAX_LSUBM+1)*5*sizeof(int) );

  // first find the \sum_i < \tilde \Psi | s_i^z | \Psi > bit.

  stagMag=0.0; 
  
  for(i=1;i<=UnitCell.number;i++) {
    subLatticeMag[i] = 0.5*UnitCell.s2[i];
    stagMag         += 0.5*UnitCell.s2[i];
  }
  
  stagMag=stagMag/((double) UnitCell.number);

  for(i=1;i<=NumberConfigurations;i++) {

    length=findClusterFromDescriptor(spinConfigurations[i].descriptor,
				     configuration);

    stagMag -= length*bra_x[i]*x[i];    
    sublattice_term(length,configuration,subTerm);
    
    for(j=1;j<=UnitCell.number;j++) 
      subLatticeMag[j] -= subTerm[j] * bra_x[i] * x[i]
                        * UnitCell.number / LatticeSymmetries.number;
  }

  free(configuration);
  return stagMag;
}

void sublattice_term(int length, int (*configuration)[5], int *subTerm)
{
  
  /* 
     This routine finds contribution to the sublattice magnetisation
     from a given cluster, namely:
     
     F(m) . S(+)(m) 
  */
  int i, j;
  static int tempConfiguration[MAX_LSUBM+1][5];
  
  for(i=1;i<=UnitCell.number;i++) subTerm[i]=0;

  // find symmetry related clusters
  for(j=1;j<=LatticeSymmetries.number;j++) { 
    /* ToDo: configuration modified ?? */
    symmetryGroup(j,length,configuration,tempConfiguration);
    for(i=1;i<=length;i++) { subTerm[tempConfiguration[i][4]]++; }
  } 

  return;
}

// new code to find the generalised expectation values.

void findGenExp(double *x, double *bra_x, double *g_exp, double *alpha)
{

  /* FIND GENERALISED EXPECTATION VALUES: obtains the staggered magnetisation.
     */

  int      i,j,index,term;
  int      (*configuration)[5];
  int      expect,error;
  int	   indices[7]={0};
  double   dummy[1],multiplier;
  double   xi2, xi3, xi4, xi5,bx;

  configuration = (int (*)[5]) malloc( (MAX_LSUBM+1)*5*sizeof(int) );
    
  for(i=0;i<=Params.nExpect;i++) {
    g_exp[i]=0.0;
  }

  term_rewind(2); // term_open(2,"r"); Jun08

  for(expect=1;expect<=Params.nExpect;expect++) {
    for (j=1;j<=2;j++) {
      for(i=1;;i++) {

        error=term_get(2,indices,dummy); /* .exp_terms */
        if (indices[0] < 0 || indices[0] > NumberConfigurations) break;
        index=indices[0];
        // the index for the bra-state correlation coefficient.
	// if this is zero then it's one of the "gs" like terms.
	
	term=indices[1];
	bx=bra_x[index];
	if (index < 1) bx=1.0;
	
	xi2=x[indices[2]];
	xi3=x[indices[3]];
	xi4=x[indices[4]];
	xi5=x[indices[5]];

        multiplier=1.0;

#if UseExpAlpha
	multiplier=usr_exp_angle(expect,term,alpha,UseAlpha);
#endif
			
	if (Control.output > 3 && myid ==0) 
	  printf("expect=%d : config=%d : term =%d : (%d,%d,%d,%d) : mult1=%10.6f : mult2=%10.6f : g_exp=%10.6f\n",
	         expect,indices[0],indices[1],indices[2],indices[3],indices[4],indices[5],
		 dummy[0],multiplier,g_exp[expect]);

        g_exp[expect] += bx*xi2*xi3*xi4*xi5*dummy[0]*multiplier;
			
      }
    }
  }

  // fclose(fp_exp); Jun08
  free(configuration);
  return;
}

// DJJF Jun08: new code to find the excitation energy values.
// Joerg: add MPI functionality

double findExcitations(double *x, double *bra_x, double *alpha)
{

  /* Find the lowest lying excitation energies (ee) from the CCM excited-state equations.

     Solved as an eigenvalue problem using the eispack routine rg_().
	 
	 Written by Damian JJ Farnell: 27 June 2008. 
	 
	 Copyright: Damian JJ Farnell (June 2008). All rights reserved.
     */

  int      i,j,index,term,bonds;
  int      term_get_error;
  int	   indices[7]={0};
  double   dummy[1];
  double   xi2, xi3, xi4, xi5;
  double   multiplier1,multiplier2;
  double   factorial[100];  
  int length1,length2,length3,length4,length5;
  int (*tempConfig)[5]=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));  
  int term2;

  factorial[0]=1.0;	
  factorial[1]=1.0;
  for(i=2;i<100;i++) factorial[i]=factorial[i-1]*((double) i);
	
  double ee=0.0;
    
  // now open the excitations CCM equations file.

  term_rewind(3);	// .exc_terms
//  char name[256];
//  strcpy(name,Control.name); 
//  strcat(name,".exc_terms");
//  fp_exc=fopen(name,"r");
//  if (!fp_exc) {
//    printf("Cannot Open %s...\nExiting!!\n\n", name);
//    exit(1);
//  }
  
  // find the initial v
  
  static double *m=NULL;  // matrix
  // if (NumberExcitedConfigurations > 1) 
  // Joerg Jun08: needs much memory cfg^2, ToDo: split into threads!
  {
    if (!m) // malloc once, never free
    m=dvector(0,(NumberExcitedConfigurations)
               *(NumberExcitedConfigurations));
    if (!m) { printf("ERROR: ecs_matrix=malloc(numCfg^2)\n"); return 0.0; }
    for (i=0;i<NumberExcitedConfigurations
              *NumberExcitedConfigurations;i++) m[i]=0.0;
  }
  ee=0.0;
  
  for(bonds=1;bonds<=Params.bonds;bonds++) {
    for(i=1;;i++) {
      term_get_error=term_get(3,indices,dummy); // .exc_terms
      if (indices[0]<0) break;
			
      if (indices[0] <= 0 || indices[0] > NumberExcitedConfigurations) {
        printf("Something funny: index[0] =%d which is <= 0 or > num exc configs!\n",indices[0]);
	exit(0);
      }
			
      index=indices[0]; // the index for the excited correlation coefficient.
			// This is never zero for excited states	
      multiplier1=1.0;
			
      term=indices[1]; // this is the term number (of 81) in the excited-state equations  

      xi3=x[indices[3]];    // a CCM correlation coefficient
      xi4=x[indices[4]];    // a CCM correlation coefficient
      xi5=x[indices[5]];    // a CCM correlation coefficient
			
      if (indices[3] < 1) xi3=1.0;
      if (indices[4] < 1) xi4=1.0;
      if (indices[5] < 1) xi5=1.0;

      // term2=mapping[term];  // map 1..81 => 1..62 [ => 1..13]
      // Joerg Jun08: mapping done in obtainExc... now
      term2=term;

      multiplier1=delta[bonds][term2];

#if UseAlpha
      multiplier1=multiplier1*usr_angle(bonds/*expect*/,term,alpha,UseAlpha);
#endif
				
      length1=findClusterFromDescriptor(
         excitedSpinConfigurations[indices[0]].descriptor, tempConfig);
									   
      length2=findClusterFromDescriptor(
         excitedSpinConfigurations[indices[2]].descriptor, tempConfig);
									   									   
      multiplier2=factorial[length2]/factorial[length1]; 
         // THIS IS A POTENTIAL PROBLEM!!
         // not at all sure about this factor!
         // However, it seems to work......
      
//      if (NumberExcitedConfigurations==1) {
//        ee+=xi3*xi4*xi5*dummy[0]*multiplier1*multiplier2;
//      } else 
      {
        // x[2] <- excitation coefficients
        // x[3] to x[5] <= g-s ket-state coefficients (NO PARTICULAR ORDER)
        index=(indices[0]-1)*NumberExcitedConfigurations
              +indices[2]-1;
	if (index >=  NumberExcitedConfigurations
                     *NumberExcitedConfigurations) {
	   printf("Index > ne*ne!\nExiting..\n\n"); ERR(1);
        }
	m[index]+=xi3*xi4*xi5*dummy[0]*multiplier1*multiplier2;
      }
			
      if (Control.output > 3) 
        printf("Bond=%d : config=%d : term =%d : (%d) : (%d,%d,%d) : mult1=%10.6f : mult2=%10.6f : mult3=%10.6f : tot=%10.6f\n",
          bonds,indices[0],indices[1],indices[2],indices[3],indices[4],indices[5],
	  dummy[0],multiplier1,multiplier2,xi3*xi4*xi5*dummy[0]*multiplier1*multiplier2);
    }
  }

  // fclose(fp_exc);

#if (PARALLEL_FLAG> 0)
  // if (NumberExcitedConfigurations > 1) 
  {
    static double *m2=NULL;
    if (!m2)
    m2=dvector(0,(NumberExcitedConfigurations)
                *(NumberExcitedConfigurations));
    for (i=0;i<NumberExcitedConfigurations
              *NumberExcitedConfigurations;i++) m2[i]=m[i];
    MPI_Allreduce(m2, m, NumberExcitedConfigurations
                        *NumberExcitedConfigurations,
                  MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    free_dvector(m2,0,NumberExcitedConfigurations
                     *NumberExcitedConfigurations);
    // now we can use single-task algorithm to go further
  }
#endif


  if (NumberExcitedConfigurations>1) {  // non-parallel

    double *wi,*wr,*z,*fv1;
    long int *iv1;

    // ToDo: Jun08, alpha version had a memory leak!
    wi =dvector(0,(NumberExcitedConfigurations)); 
    wr =dvector(0,(NumberExcitedConfigurations)); 
    iv1=(int long *)malloc(2*NumberExcitedConfigurations*sizeof(long int));
    if (!iv1) ERR("malloc iv1[2*NumExcCfgs]");
    fv1=dvector(0,(2*NumberExcitedConfigurations)); 
    z  =dvector(0,(2*NumberExcitedConfigurations));
   
    long int order1=0;
    long int    matz=0;
    long int    ierr=0;
    int stat;

    order1=(long) NumberExcitedConfigurations;

// DJJF test: matrix LSUB4 Heisenberg square from old CCM code
// Joerg: hmm, I have other results for LSUB4 Heisenberg square
/* 6*6 matrix:
 *m= { 1.122, -12.000, -12.000, -24.000, -24.000, -24.000, 
      -0.048,   4.982,   0.131,  -0.046,  -0.092,  -0.033, 
       0.005,   0.075,   5.985,  -0.006,  -0.052,  -0.350, 
      -0.038,  -0.023,  -0.017,   5.270,  -0.367,  -0.102, 
       0.004,   0.000,  -0.020,  -0.160,   5.963,  -0.195, 
       0.009,   0.003,  -0.175,  -0.005,  -0.195,   5.878};  // OK 
*/
// DJJF: Gives the correct value of "Excited-State Energy = 0.8497504284"
// JS:   but LSUB4 value and eigenvalue(m) is: 0.851867601151 !!! (see papers)
//       inaccuracy? because we only have %.3f here?

    if (Control.output > 3 && myid==0) {
      printf("\nCCM Excited-State Eigenvalue Matrix To Be Diagonalised = \n");
      for (i=0; i<order1; i++) {    
        for (j=0; j<order1; j++) {    
          printf(" % 8.4f",m[i*NumberExcitedConfigurations+j]);
        }
        printf("\n");
      }
      printf("\n");
    }

    stat=rg_(&order1 , &order1 , m ,wr , wi, &matz, z, iv1, fv1, &ierr);
      // eispack eigenvalue routine
		
    if (ierr !=0) 
      printf("Ierror in solve excitations = %d\n",(int) ierr);
   
    ee=1e10; // ee contains the lowest-valued eigenvalue
    for (i=0; i<order1; i++) {    
      if (Control.output > 2 && myid==0) {
        printf("Excited-state Eigenvalue (%d) = ( %16.12f , %16.12f ) \n",
               i+1,wr[i], wi[i]);
      }
      if (wr[i] <ee) ee=wr[i];
    }
    free_dvector(wi, 0,(NumberExcitedConfigurations));
    free_dvector(wr, 0,(NumberExcitedConfigurations));
    free(iv1);
    free_dvector(fv1,0,(2*NumberExcitedConfigurations));
    free_dvector(z,  0,(2*NumberExcitedConfigurations));
  } else ee=m[0];

  return ee;
}
